package net.wanji.opt.service.impl;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.date.DateUnit;
import cn.hutool.core.date.DateUtil;
import net.wanji.common.dto.CrossIdDTO;
import net.wanji.common.enums.CrossStatusEnum;
import net.wanji.common.enums.EventAlarmEnum;
import net.wanji.common.enums.EventAlarmSourceEnum;
import net.wanji.common.utils.tool.CrossUtil;
import net.wanji.databus.dao.entity.RidInfoEntity;
import net.wanji.databus.dao.mapper.CrossDataRealtimeMapper;
import net.wanji.databus.dao.mapper.CrossDirDataHistMapper;
import net.wanji.databus.dao.mapper.CrossTurnDataRealtimeMapper;
import net.wanji.databus.dao.mapper.RidInfoMapper;
import net.wanji.databus.po.CrossDataRealtimePO;
import net.wanji.databus.po.CrossDirDataHistPO;
import net.wanji.databus.po.CrossTurnDataRealtimePO;
import net.wanji.databus.vo.AbnormalCrossListVO;
import net.wanji.opt.dao.mapper.CrossDirInfoMapper;
import net.wanji.opt.dao.mapper.CrossInfoMapper;
import net.wanji.opt.dao.mapper.trend.EventAlarmMapper;
import net.wanji.opt.dao.mapper.trend.GreenwaveInfoMapper;
import net.wanji.opt.dto.trend.AbnormalCrossListDTO;
import net.wanji.opt.dto.trend.GreenwaveListDTO;
import net.wanji.opt.po.base.CrossInfoPO;
import net.wanji.opt.po.trend.EventAlarmPO;
import net.wanji.opt.service.TrendService;
import net.wanji.opt.vo.AbnormalCrossDetailVO;
import net.wanji.opt.vo.AbnormalCrossStats;
import net.wanji.opt.vo.AbnormalCrossVO;
import net.wanji.opt.vo.EventAlarmVO;
import net.wanji.opt.vo.GreenwaveListVO;
import net.wanji.opt.vo.GreenwaveStats;
import org.springframework.stereotype.Service;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * @author Kent HAN
 * @date 2023/2/8 10:31
 */
@Service
public class TrendServiceImpl implements TrendService {
    private final GreenwaveInfoMapper greenwaveInfoMapper;
    private final CrossInfoMapper crossInfoMapper;
    private final CrossDataRealtimeMapper crossDataRealtimeMapper;
    private final CrossTurnDataRealtimeMapper crossTurnDataRealtimeMapper;
    private final CrossDirDataHistMapper crossDirDataHistMapper;
    private final EventAlarmMapper eventAlarmMapper;
    private final RidInfoMapper ridInfoMapper;
    private final CrossDirInfoMapper crossDirInfoMapper;

    SimpleDateFormat sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");

    public TrendServiceImpl(GreenwaveInfoMapper greenwaveInfoMapper, CrossInfoMapper crossInfoMapper,
                            CrossDataRealtimeMapper crossDataRealtimeMapper,
                            CrossTurnDataRealtimeMapper crossTurnDataRealtimeMapper,
                            CrossDirDataHistMapper crossDirDataHistMapper, EventAlarmMapper eventAlarmMapper,
                            RidInfoMapper ridInfoMapper, CrossDirInfoMapper crossDirInfoMapper) {
        this.greenwaveInfoMapper = greenwaveInfoMapper;
        this.crossInfoMapper = crossInfoMapper;
        this.crossDataRealtimeMapper = crossDataRealtimeMapper;
        this.crossTurnDataRealtimeMapper = crossTurnDataRealtimeMapper;
        this.crossDirDataHistMapper = crossDirDataHistMapper;
        this.eventAlarmMapper = eventAlarmMapper;
        this.ridInfoMapper = ridInfoMapper;
        this.crossDirInfoMapper = crossDirInfoMapper;
    }

    @Override
    public List<GreenwaveListVO> greenwaveList(GreenwaveListDTO greenwaveListDTO) {
        Integer status = greenwaveListDTO.getStatus();
        String name = greenwaveListDTO.getName();
        Integer type = greenwaveListDTO.getType();
        List<GreenwaveListVO> greenwaveListVOList = greenwaveInfoMapper.listGreenwave(status, name, type);
        // 3、4都算拥堵
        if (status != null && status == 3) {
            List<GreenwaveListVO> extraList = greenwaveInfoMapper.listGreenwave(4, name, type);
            greenwaveListVOList.addAll(extraList);
        }
        changeStatus(greenwaveListVOList);
        for (GreenwaveListVO greenwaveListVO : greenwaveListVOList) {
            // 计算持续时间
            buildDuration(greenwaveListVO);
            // 构造路口列表
            String greenwaveId = greenwaveListVO.getId();
            List<GreenwaveListVO.CrossListElement> crossList = buildCrossList(greenwaveId);
            greenwaveListVO.setCrossList(crossList);
        }
        return greenwaveListVOList;
    }

    @Override
    public GreenwaveStats buildGreenwaveStats(List<GreenwaveListVO> greenwaveListVOList) {
        GreenwaveStats greenwaveStats = new GreenwaveStats();
        Integer smooth = greenwaveStats.getSmooth();
        Integer slow = greenwaveStats.getSlow();
        Integer congestion = greenwaveStats.getCongestion();
        for (GreenwaveListVO greenwaveListVO : greenwaveListVOList) {
            Integer realtimeStatus = greenwaveListVO.getRealtimeStatus();
            if (realtimeStatus == 1) {
                smooth += 1;
            } else if (realtimeStatus == 2) {
                slow += 1;
            } else if (realtimeStatus == 3) {
                congestion += 1;
            }
        }
        greenwaveStats.setSmooth(smooth);
        greenwaveStats.setSlow(slow);
        greenwaveStats.setCongestion(congestion);
        return greenwaveStats;
    }

    @Override
    public AbnormalCrossVO abnormalCrossList(AbnormalCrossListDTO abnormalCrossListDTO) {
        AbnormalCrossVO abnormalCrossVO = new AbnormalCrossVO();
        // 构建问题路口列表
        Integer status = abnormalCrossListDTO.getStatus();
        String name = abnormalCrossListDTO.getName();
        Integer type = abnormalCrossListDTO.getType();
        List<AbnormalCrossListVO> abnormalCrossListVOList =
                crossDataRealtimeMapper.selectAbnormalCross(status, name, type);
        // 坐标格式转换
        for (AbnormalCrossListVO abnormalCrossListVO : abnormalCrossListVOList) {
            String locationStr = abnormalCrossListVO.getLocationStr();
            double[] lonLat = CrossUtil.getLonLat(locationStr);
            List<Double> location = new ArrayList<>();
            location.add(lonLat[0]);
            location.add(lonLat[1]);
            abnormalCrossListVO.setLocation(location);
            // 关联路口信息
            Integer realtimeStatus = abnormalCrossListVO.getRealtimeStatus();
            String endCrossId = abnormalCrossListVO.getId();
            if (Objects.equals(realtimeStatus, CrossStatusEnum.SPILLOVER.getCode())) {
                // 溢出只返回溢出方向的关联路口
                String spilloverDirs = abnormalCrossListVO.getSpilloverDirs();
                String[] split = spilloverDirs.split(",");
                List<Integer> spilloverDirList = new ArrayList<>();
                for (String s : split) {
                    spilloverDirList.add(Integer.parseInt(s));
                }
                buildRidData(abnormalCrossListVO, endCrossId, spilloverDirList);
            } else if (Objects.equals(realtimeStatus, CrossStatusEnum.CONGESTION.getCode())){
                // 拥堵返回所有方向关联路口
                List<Integer> inDirs = crossDirInfoMapper.selectInDirsByCrossId(endCrossId);
                buildRidData(abnormalCrossListVO, endCrossId, inDirs);
            }
        }
        CrossComparator crossComparator = new CrossComparator();
        abnormalCrossListVOList.sort(crossComparator);
        abnormalCrossVO.setAbnormalCrossList(abnormalCrossListVOList);
        // 构造统计信息
        List<AbnormalCrossListVO> listForStats =
                crossDataRealtimeMapper.selectAbnormalCross(null, null, null);
        buildAbnormalCrossStats(abnormalCrossVO, listForStats);

        return abnormalCrossVO;
    }

    private static class CrossComparator implements Comparator<AbnormalCrossListVO> {
        @Override
        public int compare(AbnormalCrossListVO o1, AbnormalCrossListVO o2) {
            // 先按路口状态排序
            Integer o1status = o1.getRealtimeStatus();
            Integer o2status = o2.getRealtimeStatus();
            int flag = o2status - o1status;
            if (flag == 0) {
                // 再按拥堵时长排序
                Long o1duration = o1.getDuration();
                Long o2Duration = o2.getDuration();
                long res = o2Duration - o1duration;
                return (int) res;
            }
            return flag;
        }
    }

    private void buildRidData(AbnormalCrossListVO abnormalCrossListVO, String endCrossId,
                              List<Integer> spilloverDirList) {
        List<List<Double>> ridLocationList = new ArrayList<>();
        List<String> ridWktList = new ArrayList<>();
        for (Integer spilloverDir : spilloverDirList) {
            RidInfoEntity ridInfo = ridInfoMapper.selectByEndInDir(endCrossId, spilloverDir);
            if (ridInfo != null) {
                // 关联路口坐标
                String startCrossId = ridInfo.getStartCrossId();
                CrossInfoPO crossInfoPO = crossInfoMapper.selectById(startCrossId);
                if (crossInfoPO != null) {
                    String startLocationStr = crossInfoPO.getLocation();
                    double[] startLonLat = CrossUtil.getLonLat(startLocationStr);
                    List<Double> startLocation = new ArrayList<>();
                    startLocation.add(startLonLat[0]);
                    startLocation.add(startLonLat[1]);
                    ridLocationList.add(startLocation);
                }
                // 关联路口wkt
                String wkt = ridInfo.getWkt();
                ridWktList.add(wkt);
            }
        }
        abnormalCrossListVO.setRidLocationList(ridLocationList);
        abnormalCrossListVO.setRidWktList(ridWktList);
    }

    @Override
    public AbnormalCrossDetailVO abnormalCrossDetail(CrossIdDTO crossIdDTO) {
        String crossId = crossIdDTO.getCrossId();
        AbnormalCrossDetailVO abnormalCrossDetailVO = new AbnormalCrossDetailVO();

        // 构造路口实时数据
        CrossDataRealtimePO crossDataRealtimePO = crossDataRealtimeMapper.selectByCrossId(crossId);
        abnormalCrossDetailVO.setCrossStatus(crossDataRealtimePO.getStatus());
        abnormalCrossDetailVO.setCrossQueueLength(crossDataRealtimePO.getQueueLength());
        abnormalCrossDetailVO.setCrossFlowRate((double)crossDataRealtimePO.getFlow());
        abnormalCrossDetailVO.setCrossStopTimes(crossDataRealtimePO.getStopTimes());
        abnormalCrossDetailVO.setCongestionIndex(crossDataRealtimePO.getCongestionIndex());
        // 构造转向实时数据
        List<AbnormalCrossDetailVO.TurnDataElement> turnData = buildTurnData(crossId);
        abnormalCrossDetailVO.setTurnData(turnData);
        // 构造方向历史数据
        List<AbnormalCrossDetailVO.DirDataElement> dirData = buildDirData(crossId);
        abnormalCrossDetailVO.setDirData(dirData);

        return abnormalCrossDetailVO;
    }

    @Override
    public List<EventAlarmVO> eventAlarm(Date todayTime) throws ParseException {
        List<EventAlarmPO> eventAlarmPOList = eventAlarmMapper.selectByTodayTime(todayTime);
        List<EventAlarmVO> eventAlarmOutVOList = new ArrayList<>();
        for (EventAlarmPO eventAlarmPO : eventAlarmPOList) {
            EventAlarmVO eventAlarmOutVO = new EventAlarmVO();
            eventAlarmOutVO.setEventDesc(eventAlarmPO.getEventDesc());
            // 事件类型
            Integer typeCode = eventAlarmPO.getType();
            String msgByCode = EventAlarmEnum.getMsgByCode(typeCode);
            eventAlarmOutVO.setType(msgByCode);

            Date endTime = eventAlarmPO.getEndTime();
            if (typeCode == 1 || typeCode == 2) {
                // 发生时间
                Date startTime = eventAlarmPO.getStartTime();
                String startTimeStr = sdf.format(startTime);
                String substring = startTimeStr.substring(11, 16);
                eventAlarmOutVO.setStartTime(substring);
                // 时长
                if (endTime == null) {
                    Date nowTime = DateUtil.date();
                    long between = DateUtil.between(startTime, nowTime, DateUnit.MINUTE);
                    int betweenInt = (int) between;
                    eventAlarmOutVO.setDuration(betweenInt);
                } else {
                    long between = DateUtil.between(startTime, endTime, DateUnit.MINUTE);
                    int betweenInt = (int) between;
                    eventAlarmOutVO.setDuration(betweenInt);
                }
            }
            if (typeCode == 3 || typeCode == 4) {
                // 时间范围
                Date startTime = eventAlarmPO.getStartTime();
                String startTimeStr = sdf.format(startTime);
                String startTimeSubStr = startTimeStr.substring(0, 16); // 截取到分钟
                String endTimeStr = sdf.format(endTime);
                String endTimeSubStr = endTimeStr.substring(0, 16); // 截取到分钟
                eventAlarmOutVO.setTimeSpan(startTimeSubStr + "~" + endTimeSubStr);
            }

            Integer sourceCode = eventAlarmPO.getSource();
            String sourceStr = EventAlarmSourceEnum.getMsgByCode(sourceCode);
            eventAlarmOutVO.setSource(sourceStr);
            eventAlarmOutVO.setCoordinate(eventAlarmPO.getWkt());

            eventAlarmOutVOList.add(eventAlarmOutVO);

        }
        return eventAlarmOutVOList;
    }

    private List<AbnormalCrossDetailVO.DirDataElement> buildDirData(String crossId) {
        List<AbnormalCrossDetailVO.DirDataElement> dirData = new ArrayList<>();
        // 获取当前时间之前40分钟的10位时间戳
        long currentSeconds = DateUtil.currentSeconds();
        long preSeconds = currentSeconds - 40 * 60;
        List<CrossDirDataHistPO> crossDirDataHistPOList =
                crossDirDataHistMapper.selectByCrossIdAndTimestamp(crossId, preSeconds);

        for (CrossDirDataHistPO crossDirDataHistPO : crossDirDataHistPOList) {
            AbnormalCrossDetailVO.DirDataElement dirDataElement = new AbnormalCrossDetailVO.DirDataElement();
            dirDataElement.setDir(crossDirDataHistPO.getDirType());
            // 获取采集时间
            Integer batchTime = crossDirDataHistPO.getBatchTime();
            Date date = DateUtil.date(batchTime * 1000);
            dirDataElement.setTime(date);
            // 转换状态，3、4都算拥堵
            Integer status = crossDirDataHistPO.getStatus();
            if (status == 4) {
                status = 3;
            }
            dirDataElement.setStatus(status);

            dirData.add(dirDataElement);
        }

        return dirData;
    }

    private List<AbnormalCrossDetailVO.TurnDataElement> buildTurnData(String crossId) {
        List<AbnormalCrossDetailVO.TurnDataElement> turnData = new ArrayList<>();
        List<CrossTurnDataRealtimePO> crossTurnDataRealtimePOList = crossTurnDataRealtimeMapper.selectByCrossId(crossId);
        Map<Integer, List<CrossTurnDataRealtimePO>> collect = crossTurnDataRealtimePOList.stream()
                .collect(Collectors.groupingBy(CrossTurnDataRealtimePO::getInDir));
        for (Map.Entry<Integer, List<CrossTurnDataRealtimePO>> entry : collect.entrySet()) {
            AbnormalCrossDetailVO.TurnDataElement turnDataElement = new AbnormalCrossDetailVO.TurnDataElement();
            Integer dir = entry.getKey();
            turnDataElement.setDir(dir);
            // 构造转向列表
            List<CrossTurnDataRealtimePO> subList = entry.getValue();
            List<AbnormalCrossDetailVO.TurnDataElement.TurnListElement> turnList = buildTurnList(subList);
            turnDataElement.setTurnList(turnList);
            
            turnData.add(turnDataElement);
        }
        return turnData;
    }

    private List<AbnormalCrossDetailVO.TurnDataElement.TurnListElement> buildTurnList(
            List<CrossTurnDataRealtimePO> subList) {
        List<AbnormalCrossDetailVO.TurnDataElement.TurnListElement> turnList = new ArrayList<>();

        for (CrossTurnDataRealtimePO crossTurnDataRealtimePO : subList) {
            AbnormalCrossDetailVO.TurnDataElement.TurnListElement turnListElement =
                    new AbnormalCrossDetailVO.TurnDataElement.TurnListElement();

            turnListElement.setTurn(crossTurnDataRealtimePO.getTurnType());
            turnListElement.setTurnQueueLength(crossTurnDataRealtimePO.getQueueLength());
            turnListElement.setTurnStopTimes(crossTurnDataRealtimePO.getStopTimes());
            turnListElement.setFlow(crossTurnDataRealtimePO.getFlow());

            turnList.add(turnListElement);
        }
        return turnList;
    }

    private static void buildAbnormalCrossStats(AbnormalCrossVO abnormalCrossVO,
                                                List<AbnormalCrossListVO> abnormalCrossListVOList) {
        AbnormalCrossStats abnormalCrossStats = new AbnormalCrossStats();
        Integer normal = abnormalCrossStats.getNormal();
        Integer unbalance = abnormalCrossStats.getUnbalance();
        Integer congestion = abnormalCrossStats.getCongestion();
        Integer spillover = abnormalCrossStats.getSpillover();
        for (AbnormalCrossListVO abnormalCrossListVO : abnormalCrossListVOList) {
            Integer realtimeStatus = abnormalCrossListVO.getRealtimeStatus();
            if (realtimeStatus == 0) {
                normal += 1;
            } else if (realtimeStatus == 1) {
                unbalance += 1;
            } else if (realtimeStatus == 2) {
                congestion += 1;
            } else if (realtimeStatus == 3) {
                spillover += 1;
            }
            if (realtimeStatus !=1 && abnormalCrossListVO.getIsUnbalance() == 1) {
                unbalance += 1;
            }
            if (realtimeStatus !=2 && abnormalCrossListVO.getIsCongestion() ==1) {
                congestion += 1;
            }
            if (realtimeStatus !=3 && abnormalCrossListVO.getIsSpillover() == 1) {
                spillover += 1;
            }
        }
        abnormalCrossStats.setNormal(normal);
        abnormalCrossStats.setUnbalance(unbalance);
        abnormalCrossStats.setCongestion(congestion);
        abnormalCrossStats.setSpillover(spillover);
        abnormalCrossVO.setAbnormalCrossStats(abnormalCrossStats);
    }

    private static void changeStatus(List<GreenwaveListVO> greenwaveListVOList) {
        if (greenwaveListVOList != null) {
            for (GreenwaveListVO greenwaveListVO : greenwaveListVOList) {
                if (greenwaveListVO.getRealtimeStatus() == 4) {
                    greenwaveListVO.setRealtimeStatus(3);
                }
            }
        }
    }

    private List<GreenwaveListVO.CrossListElement> buildCrossList(String greenwaveId) {
        List<GreenwaveListVO.CrossListElement> crossList = new ArrayList<>();
        List<String> crossIdList = greenwaveInfoMapper.selectCrossIdsById(greenwaveId);
        if (CollectionUtil.isNotEmpty(crossIdList)) {
            List<CrossInfoPO> crossInfoPOList = crossInfoMapper.selectByCrossIds(crossIdList);
            for (CrossInfoPO crossInfoPO : crossInfoPOList) {
                GreenwaveListVO.CrossListElement crossListElement = new GreenwaveListVO.CrossListElement();
                crossListElement.setCrossId(crossInfoPO.getId());
                crossListElement.setName(crossInfoPO.getName());
                List<Double> location = buildLocation(crossInfoPO);
                crossListElement.setLocation(location);

                crossList.add(crossListElement);
            }
        }
        return crossList;
    }

    private List<Double> buildLocation(CrossInfoPO crossInfoPO) {
        String location = crossInfoPO.getLocation();
        double[] lonLat = CrossUtil.getLonLat(location);
        List<Double> res = new ArrayList<>();
        res.add(lonLat[0]);
        res.add(lonLat[1]);
        return res;
    }

    private static void buildDuration(GreenwaveListVO greenwaveListVO) {
        Date startTime = greenwaveListVO.getStartTime();
        Date endTime = greenwaveListVO.getEndTime();
        Long duration = null;
        if (endTime != null) {
            duration = DateUtil.between(startTime, endTime, DateUnit.MINUTE);
        } else {
            duration = DateUtil.between(startTime, new Date(), DateUnit.MINUTE);
        }
        greenwaveListVO.setDuration(duration);
    }
}
